/****************************************************************************************/
/* Basis.ino                                                                            */
/* Collection of basis routines                                                         */
/*                                                                                      */
/* 1. POWER ON/OFF routines                                                             */
/* SUB_POWER_OFF:           Subroutine to switch device in power off state              */
/* SUB_POWER_ON:            Subroutine to switch device in power on state               */
/*                                                                                      */
/* 2. FLASH routines:                                                                   */
/* SUB_INIT_FLASH:          Subroutine to initialize FLASH with default values          */
/*                                                                                      */
/* 3. INIT FILE routines:                                                               */
/* SUB_READ_INIT_FILE:      Subroutine to read out init file Init.txt                   */
/*                                                                                      */
/* 4. DEBUG routines:                                                                   */
/* SUB_DISPLAY_FLASH:       Subroutine to display all FLASH values                      */
/*                                                                                      */
/* 5. NEOPIXEL routines:                                                                */
/* SUB_FAST_LED_SHOW_TASK;  Task handler for FastLED library                            */
/* SUB_NEOPIXEL_SEND:       Subroutine to adjust Neopixel colors                        */
/****************************************************************************************/


/****************************************************************************************/
/****************************************************************************************/
/* POWER ON/OFF ROUTINES                                                                */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine to switch device in power off state                                       */
/****************************************************************************************/ 
void SUB_POWER_OFF() 
{
  #ifdef _DEBUG_
    Serial.println ("Switching power off ..."); 
  # endif 
  SUB_NEOPIXEL_SEND (0,0,0);                      /* Switch all colors off */ 
  #ifdef _DEBUG_
    Serial.print ("Saving actual program: ");
    Serial.println (ACT_PROGRAM);
  # endif 
  EEPROM.write(0, ACT_PROGRAM);                   /* Save actual program in FLASH */
  EEPROM.commit();                                /* Store values permanently in FLASH */
  /* Single color selected ? */
  if ((ACT_PROGRAM >=1) && (ACT_PROGRAM <= 13))   /* Program for single color? */
  {
    #ifdef _DEBUG_
      Serial.print ("Saving actual colour dimming value: ");
      Serial.println (COLOR_DIM_STEP);
    # endif
    EEPROM.write(1, COLOR_DIM_STEP);              /* Save actual color dimming value in FLASH */
    EEPROM.commit();                              /* Store values permanently in FLASH */
  }     
  /* Program selected ? */
  if ((ACT_PROGRAM >= 20) && (ACT_PROGRAM <= 29))           
  {
    #ifdef _DEBUG_
      Serial.print ("Saving actual program dimming value: ");
      Serial.println (PROG_DIM_STEP);
      Serial.print ("Saving actual program speed value: ");
      Serial.println (PROG_SPEED);      
    # endif
    EEPROM.write(2, PROG_DIM_STEP);               /* Save actual program dimming value in FLASH */
    EEPROM.write(3, PROG_SPEED);                  /* Save program speed in FLASH */ 
    EEPROM.commit();                              /* Store values permanently in FLASH */
  }
  POWER_OFF = true;                               /* Set power off flag true */ 
  POWER_WEB_OFF = true;                           /* Set power off flag true for web */ 
  if (KEY_PRESSED == true)                        /* Power on via push button ? */
  {
    delay (2000);                                 /* 2 seconds debouncing delay */
    KEY_PRESSED = false;                          /* Set key pressed flag to false */
  }
  SUB_NEOPIXEL_SEND (0,0,0);                      /* Switch all colors off for second time to be absolutely sure */ 
  SELECT = 255;                                   /* Delete key buffer */
}


/****************************************************************************************/
/* Subroutine to switch device in power on state                                        */
/****************************************************************************************/
void SUB_POWER_ON() 
{
  ACT_PROGRAM    = EEPROM.read (0);               /* Read actual program out of FLASH */
  COLOR_DIM_STEP = EEPROM.read (1);               /* Read actual color dimming value out of FLASH */
  PROG_DIM_STEP  = EEPROM.read (2);               /* Read actual program dimming value out of FLASH */
  PROG_SPEED     = EEPROM.read (3);               /* Read program speed out of FLASH */
  SELECT = ACT_PROGRAM;                           /* Determine correct SELECT value */
  #ifdef _DEBUG_
    Serial.println ("Switching power on...");
    Serial.print ("Actual program: ");
    Serial.println (ACT_PROGRAM);
  # endif  
  POWER_OFF = false;                              /* Set power off flag false */
  POWER_WEB_OFF = false;                          /* Set power off flag for web false */
    
  if ((SELECT >=1) && (SELECT <= 13))             /* Program for single color? */
  {
    SUB_MAIN_COLOR();                             /* Call subroutine for single color */
    SELECT = 255;                                 /* Delete key buffer */
  }
    
  if ((SELECT >= 15) && (SELECT <= 19))           /* Program for stored values? */
  {
    SUB_RCL_STORAGE();                            /* Call subroutine for single storage */  
    SELECT = 255;                                 /* Delete key buffer */ 
  }

  if ((SELECT >= 20) && (SELECT <= 29))           /* Program for program sequence? */
  {
    SUB_PROG_SEL ();                              /* Call subroutine for animated programs */
    SELECT = 255;                                 /* Delete key buffer */
  }

  if (KEY_PRESSED == true)                        /* Power on via push button ? */
  {
    delay (2000);                                 /* 2 seconds debouncing delay */
    KEY_PRESSED = false;                          /* Set key pressed flag to false */
  }
}


/****************************************************************************************/
/****************************************************************************************/
/* FLASH ROUTINES                                                                       */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Initialize FLASH with default values                                                 */
/****************************************************************************************/
void SUB_INIT_FLASH (void)
{
  #ifdef _DEBUG_
    Serial.println ("Initialisation of FLASH memory ...");  
  # endif 
  EEPROM.write(0, 1);                             /* Storage position 0: Color 1 */ 
  EEPROM.write(1, 0);                             /* No dimming color */ 
  EEPROM.write(2, 0);                             /* No dimming program */ 
  EEPROM.write(3, 4);                             /* Program speed = 4 */        
  EEPROM.write(10, 144);                          /* Storage position 1: Red color value */ 
  EEPROM.write(11, 72);                           /* Storage position 1: Green color value */ 
  EEPROM.write(12, 18);                           /* Storage position 1: Blue color value */ 
  EEPROM.write(13,0);                             /* Storage position 1: Dimming factor = 0 */  
  EEPROM.write(20, 252);                          /* Storage position 2: Red color value */
  EEPROM.write(21, 45);                           /* Storage position 2: Green color value */
  EEPROM.write(22, 0);                            /* Storage position 2: Blue color value */
  EEPROM.write(23,0);                             /* Storage position 2: Dimming factor = 0 */ 
  EEPROM.write(30, 72);                           /* Storage position 3: Red color value */
  EEPROM.write(31, 45);                           /* Storage position 3: Green color value */
  EEPROM.write(32, 0);                            /* Storage position 3: Blue color value */
  EEPROM.write(33,0);                             /* Storage position 3: Dimming factor = 0 */ 
  EEPROM.write(40, 112);                          /* Storage position 4: Red color value */
  EEPROM.write(41, 0);                            /* Storage position 4: Green color value */
  EEPROM.write(42, 56);                           /* Storage position 4: Blue color value */
  EEPROM.write(43,0);                             /* Storage position 4: Dimming factor = 0 */ 
  EEPROM.write(50, 27);                           /* Storage position 5: Red color value */
  EEPROM.write(51, 99);                           /* Storage position 5: Green color value */
  EEPROM.write(52, 252);                          /* Storage position 5: Blue color value */
  EEPROM.write(53,0);                             /* Storage position 5: Dimming factor = 0 */  
  EEPROM.commit();                                /* Write to FLASH memory */
}


/****************************************************************************************/
/****************************************************************************************/
/* INIT FILE ROUTINES                                                                   */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine to read out init file Init.txt                                            */
/****************************************************************************************/
void SUB_READ_INIT_FILE (void)
{
  boolean BREAK_LOOP;                             /* Break loop flag: false = No break, true = Break */
  char RECEIVE_CHAR;                              /* Received byte from file reading */
  byte LINE_COUNTER;                              /* Line counter */
  byte IP_ADDRESS_COUNTER;                        /* IP adress counter */
  int TEST_INT;                                   /* Test integer */
  String IP_ADRESS_STRING;                        /* IP Adressstring */

  #ifdef _DEBUG_
    Serial.print("  Open file Init.txt: ");
  # endif 

  BREAK_LOOP = false;                             /* Set break loop false */
  File file = SPIFFS.open("/Init.txt", FILE_READ);/* Open Init file for read */
  if(!file)                                       /* File does not exist? */
  {
    #ifdef _DEBUG_
      Serial.println ("Failed! File does not exist");
    # endif
    BREAK_LOOP = true;                            /* Set break loop true */ 
  }
  
  if (BREAK_LOOP == false)                        /* No error yet */
  {
    #ifdef _DEBUG_
      Serial.println ("Done");
      Serial.println ("  Read Init.txt content: ");
    # endif  
    LINE_COUNTER = 1;                             /* Set line counter to first line */
    LOOP_COUNTER = 0;                             /* Check for maximum received bytes to prevent buffer overflows */

    /* Delete WLAN_SSID string */
    memset (WLAN_SSID, 0 , sizeof(WLAN_SSID));
    /* Delete WIFI passowrd string */
    memset (WLAN_SSID, 0 , sizeof(WLAN_PWD));    
       
    /* Read until end of file or error detected */
    while((file.available()) && (BREAK_LOOP == false))                    
    {
      RECEIVE_CHAR = file.read();                 /* Read char byte from file */

      if (RECEIVE_CHAR == '\r')                   /* Carrage return detected ? */
      {
        /* Fixed IP address */
        if (LINE_COUNTER == 3)                    /* Fixed IP address ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    Fixed IP-Address: ");
          # endif 
          TEST_INT = IP_ADRESS_STRING.toInt();
          /* In range ? */
          if ((TEST_INT >= 0) && (TEST_INT <= 255))
          {
            FIXED_IP[3] = TEST_INT;               /* Save 3rd digit */ 
            IP_ADDRESS_COUNTER = 5;               /* Set to fifth address */             
          }
          else                                    /* Not in range ? */
          {
            LOOP_COUNTER = 50;                    /* Set loop counter out of range */
          }
          /* Check if 4 digits received and loop counter in range */
          if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
          {
            #ifdef _DEBUG_
              Serial.println (FIXED_IP);
            # endif 
            LINE_COUNTER = 4;                     /* Set line counter to forth line */
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            LOOP_COUNTER = 0;                     /* Reset loop counter */
            IP_ADRESS_STRING = "";                /* Delete string for next digit */
          }
          else                                    /* Error ? */  
          {
            #ifdef _DEBUG_
              Serial.println ("Fixed IP address format error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }
        
        /* WIFI Password */
        if (LINE_COUNTER == 2)                    /* Receive WIFI password ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    WIFI password: ");
          # endif 
          /* Check on too long WIFI password */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (WLAN_PWD);
            # endif 
            LINE_COUNTER = 3;                     /* Set line counter to third line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */     
            IP_ADDRESS_COUNTER = 1;               /* Set to first address */
            IP_ADRESS_STRING = "";                /* Delete string for next digit */       
          }
          else                                    /* WIFI password length error ? */
          {
             #ifdef _DEBUG_
              Serial.println ("WIFI password length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }
        
        /* SSID */
        if (LINE_COUNTER == 1)                    /* Receive SSID ? */
        {
          #ifdef _DEBUG_
            Serial.print ("    WIFI SSID: ");
          # endif 
          /* Check on too long SSID */
          if ((LOOP_COUNTER > 1) && (LOOP_COUNTER < 20))        
          {
            #ifdef _DEBUG_
              Serial.println (WLAN_SSID);
            # endif 
            LINE_COUNTER = 2;                     /* Set line counter to second line */
            LOOP_COUNTER = 0;                     /* Reset loop counter */            
          }
          else                                    /* SSID length error ? */
          {
             #ifdef _DEBUG_
              Serial.println ("SSID length error!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */
          }
        }   
      }
      else                                        /* No CR received ? */
      {
        if (RECEIVE_CHAR != '\n')                 /* No line feed ? */
        {
          if (LOOP_COUNTER < 20) 
          {
            if (LINE_COUNTER == 1)                /* Receive SSID ? */
            {
              /* Add received char */
              WLAN_SSID[LOOP_COUNTER] = RECEIVE_CHAR;      
            }
            if (LINE_COUNTER == 2)                /* Receive WIFI password ? */
            {
              /* Add received char */
              WLAN_PWD[LOOP_COUNTER] = RECEIVE_CHAR;      
            }            
            if (LINE_COUNTER == 3)                /* Receive fixed IP address ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    FIXED_IP[2] = TEST_INT;       /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    FIXED_IP[1] = TEST_INT;       /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    FIXED_IP[0] = TEST_INT;       /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }

            if (LINE_COUNTER == 4)                /* Subnet mask ? */
            { 
              if (RECEIVE_CHAR != '.')            /* Delimiter not detected ? */
              {
                IP_ADRESS_STRING += RECEIVE_CHAR; /* Add char */
              }
              else                                /* Delimiter detected ? */
              {             
                if (IP_ADDRESS_COUNTER == 3)      /* 3rd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    SUBNET[2] = TEST_INT;         /* Save 3rd digit */               
                    IP_ADDRESS_COUNTER = 4;       /* Set to forth digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }
                if (IP_ADDRESS_COUNTER == 2)      /* 2nd digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    SUBNET[1] = TEST_INT;         /* Save 2rd digit */               
                    IP_ADDRESS_COUNTER = 3;       /* Set to third digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }                
                if (IP_ADDRESS_COUNTER == 1)      /* 1st digit ? */
                {
                  /* Convert to Int */
                  TEST_INT = IP_ADRESS_STRING.toInt();
                  /* In range ? */
                  if ((TEST_INT >= 0) && (TEST_INT <= 255))
                  {
                    SUBNET[0] = TEST_INT;         /* Save 1st digit */               
                    IP_ADDRESS_COUNTER = 2;       /* Set to second digit */
                    IP_ADRESS_STRING = "";        /* Delete string for next digit */
                  }
                  else                            /* Not in range ? */
                  {
                    LOOP_COUNTER = 50;            /* Set loop counter out of range */
                  }
                }  
              }    
            }            
                        
            LOOP_COUNTER++;                       /* Increment loop counter */
          }
          else                                    /* Length error */
          {
            #ifdef _DEBUG_
              Serial.println ("File length error!!!");
            # endif 
            BREAK_LOOP = true;                    /* Set break loop true */          
          }
        }
      }
    }
    file.close();                                 /* Close file after reading */
    
    if (BREAK_LOOP == false)
    {   
      #ifdef _DEBUG_
        Serial.print ("    Subnet mask: ");
      # endif 
      TEST_INT = IP_ADRESS_STRING.toInt();
      /* In range ? */
      if ((TEST_INT >= 0) && (TEST_INT <= 255))
      {
        SUBNET[3] = TEST_INT;                     /* Save 3rd digit */   
        IP_ADDRESS_COUNTER = 5;                   /* Set to fifth address */            
      }
      else                                        /* Not in range ? */
      {
        LOOP_COUNTER = 50;                        /* Set loop counter out of range */
      }
      /* Check if 4 digits received and loop counter in range */
      if ((LOOP_COUNTER < 20) && (IP_ADDRESS_COUNTER == 5))        
      {
        #ifdef _DEBUG_
          Serial.println (SUBNET);
        # endif 
      }
      else                                        /* Error ? */  
      {
        #ifdef _DEBUG_
          Serial.println ("Subnet mask format error!");
        # endif 
        BREAK_LOOP = true;                        /* Set break loop true */
      }
    }     
  }


  if (BREAK_LOOP == true)                         /* Error ? */
  {
    do                                            /* Endless loop until manual reset */
    {
      SUB_NEOPIXEL_SEND (255, 0, 0);              /* Switch all pixels to red color */
      delay (1000);                               /* 1000ms delay */
      SUB_NEOPIXEL_SEND (0, 0, 0);                /* Switch off all pixels */
      delay (1000);                               /* 1000ms delay */      
    }
    while (1);  
  }
}


/****************************************************************************************/
/****************************************************************************************/
/* DEBUG ROUTINES                                                                       */
/****************************************************************************************/
/****************************************************************************************/


/****************************************************************************************/
/* Display all FLASH values                                                             */
/****************************************************************************************/
void SUB_DISPLAY_FLASH (void)
{
  byte FLASH_VALUE;                               /* Reading FLASH value */
  
  delay (1);                                      /* Proforma delay */
  #ifdef _DEBUG_
    Serial.println ("List FLASH memory:"); 
    FLASH_VALUE = EEPROM.read(0);
    Serial.print ("Memory location 0: Actual program (1-13 = Colours, 15-19=Storage positions, 20-29=Programs): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(1);
    Serial.print ("Memory location 1: Actual dimming color (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);   
    FLASH_VALUE = EEPROM.read(2);
    Serial.print ("Memory location 2: Actual dimming program (0=no dimming, 4=max dimming): ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(3);
    Serial.print ("Memory location 3: Actual program speed (0=highest, 9=lowest): ");
    Serial.println (FLASH_VALUE);                 
    FLASH_VALUE = EEPROM.read(10);
    Serial.print ("Memory location 10: Colour storage position 1 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(11);
    Serial.print ("Memory location 11: Colour storage position 1 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(12);
    Serial.print ("Memory location 12: Colour storage position 1 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(13);
    Serial.print ("Memory location 13: Colour storage position 1 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(20);
    Serial.print ("Memory location 20: Colour storage position 2 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(21);
    Serial.print ("Memory location 21: Colour storage position 2 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(22);
    Serial.print ("Memory location 22: Colour storage position 2 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(23);
    Serial.print ("Memory location 23: Colour storage position 2 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(30);
    Serial.print ("Memory location 30: Colour storage position 3 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(31);
    Serial.print ("Memory location 31: Colour storage position 3 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(32);
    Serial.print ("Memory location 32: Colour storage position 3 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(33);
    Serial.print ("Memory location 33: Colour storage position 3 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(40);
    Serial.print ("Memory location 40: Colour storage position 4 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(41);
    Serial.print ("Memory location 41: Colour storage position 4 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(42);
    Serial.print ("Memory location 42: Colour storage position 4 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(43);
    Serial.print ("Memory location 43: Colour storage position 4 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);    
    FLASH_VALUE = EEPROM.read(50);
    Serial.print ("Memory location 50: Colour storage position 5 - Red value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(51);
    Serial.print ("Memory location 51: Colour storage position 5 - Green value: ");
    Serial.println (FLASH_VALUE); 
    FLASH_VALUE = EEPROM.read(52);
    Serial.print ("Memory location 52: Colour storage position 5 - Blue value: ");
    Serial.println (FLASH_VALUE);
    FLASH_VALUE = EEPROM.read(53);
    Serial.print ("Memory location 53: Colour storage position 5 - Dimming (0=no dimming, 9=max dimming): ");
    Serial.println (FLASH_VALUE);                   
  # endif  
}


/****************************************************************************************/
/****************************************************************************************/
/* NEOPIXEL routines                                                                    */
/****************************************************************************************/
/****************************************************************************************/

/****************************************************************************************/
/* Sub routine show Task                                                                */
/* This function runs on core 0 and just waits for requests to call FastLED.show()      */
/****************************************************************************************/ 
void SUB_FAST_LED_SHOW_TASK(void *pvParameters)
{
  for(;;)                                         /* Run forever */
  {
    ulTaskNotifyTake(pdTRUE, portMAX_DELAY);      /* Wait for the trigger */
    if (DISPLAY_LOCK == false)                    /* No lock ? */
    {
      FastLED.show();                             /* Do the show (synchronously)*/
    }
    xTaskNotifyGive(USER_TASK_HANDLE);            /* Notify the calling task */
  }
}

/****************************************************************************************/
/* Sub routine to display colors on Neopixels                                           */
/* COLOR_1 = Red color value                                                            */
/* COLOR_2 = Green color value                                                          */
/* COLOR_3 = Blue color value                                                           */
/*                                                                                      */
/* Call this function instead of FastLED.show(). It signals core 0 to issue a show,     */
/* then waits for a notification that it is done.                                       */
/****************************************************************************************/ 
void SUB_NEOPIXEL_SEND(byte COLOR_1, byte COLOR_2, byte COLOR_3)
{
  byte LOOP_LED_COUNTER;                          /* Loop led counter to fill array */
 
  for(LOOP_LED_COUNTER=0; LOOP_LED_COUNTER < LED_NUM_LEDS; LOOP_LED_COUNTER++) 
  {     
    /* Fill array with color values */                  
    leds[LOOP_LED_COUNTER] = CRGB( COLOR_2, COLOR_1, COLOR_3);      
  }
  if (USER_TASK_HANDLE == 0)                      /* User task handler is zero ? */                   
  {
    /* Store the handle of the current task, so that the show task can notify it when it's done */
    USER_TASK_HANDLE = xTaskGetCurrentTaskHandle();
    xTaskNotifyGive(FAST_LED_SHOW_TASK_HANDLE);   /* Trigger the show task */
    /* Wait to be notified that it's done */
    const TickType_t xMaxBlockTime = pdMS_TO_TICKS( 200 );
    ulTaskNotifyTake(pdTRUE, xMaxBlockTime);
    USER_TASK_HANDLE = 0;
  }
}
