/****************************************************************************************/
/* Neopixel Bulb for Pinecone lamp with ESP32 controlled push button and webpage        */
/* Version: 2.1                                                                         */
/* Board: ESP32 Dev Module                                                              */
/* Date: 28.12.2020                                                                     */
/* Compiled with Arduino 1.8.12                                                         */
/* ESP32 arduino extension: 1.0.4                                                       */
/*                                                                                      */
/* FLASH handling: https://randomnerdtutorials.com/esp32-flash-memory/                  */
/* ESP Async Web Server:                                                                */
/*   https://techtutorialsx.com/2017/12/01/esp32-arduino-asynchronous-http-webserver/   */
/*                                                                                      */
/* Libraries:                                                                           */
/*            ESP32 Async Web Server: https://github.com/me-no-dev/ESPAsyncWebServer    */
/*            Async TCP: https://github.com/me-no-dev/AsyncTCP                          */
/*            Fastled: https://github.com/FastLED/FastLED                               */
/*            Wifi: Part of ESP32 library                                               */
/*            SPIFFS: Part of ESP32 library                                             */
/*                                                                                      */
/* Other tools:                                                                         */
/*        ESP32 Filesystem Uploader:                                                    */
/*        https://randomnerdtutorials.com/install-esp32-filesystem-uploader-arduino-ide */
/*                                                                                      */
/* Usage of serial interfaces: Serial for debug interface                               */
/* Pin asignments:                                                                      */
/*     GPIO12 = IO2 = Neopixel control pin                                              */
/*     GPIO13 = IO15 = Pushbutton on/off                                                */
/*                                                                                      */
/* INIT.TXT: Init file with basic access parameters which must be uploaded via          */
/*           ESP32 Sketch Data Upload                                                   */
/*      File content:                                                                   */
/*           Line 1:  WIFI SSID                                                         */
/*           Line 2:  WIFI Password                                                     */
/*           Line 3:  Fixed IP address, e.g. 192.168.122.1                              */
/*           Line 4:  Subnet mask, e.g. 255.255.255.0                                   */
/*      Warning: No ENTER (CR/LF) allowed after line 4 !!!                              */  
/*                                                                                      */
/* Internal command reference                                                           */
/*                         1-13 Single color:                                           */
/*                         1 = Single color cold white                                  */
/*                         2 = Single color warm white                                  */
/*                         3 = Single color yellow                                      */
/*                         4 = Single color orange                                      */
/*                         5 = Single color red                                         */
/*                         6 = Single color dark purple                                 */
/*                         7 = Single color purple                                      */
/*                         8 = Single color bright purple                               */
/*                         9 = Single color blue                                        */
/*                        10 = Single color bright blue                                 */
/*                        11 = Single color cyan                                        */
/*                        12 = Single color bright green                                */
/*                        13 = Single color green                                       */
/*                                                                                      */
/*                        15-19 Storage position:                                       */
/*                        15 = Storage position 1                                       */
/*                        16 = Storage position 2                                       */
/*                        17 = Storage position 3                                       */
/*                        18 = Storage position 4                                       */
/*                        19 = Storage position 5                                       */
/*                                                                                      */
/*                        20-29 Programs:                                               */
/*                        20 = Program 0: Change 1: 3 Second hold, 3 Second change      */
/*                                                  White - Red - White - Green -       */
/*                                                  White - Blue - White - Yellow -     */
/*                                                  White - Cyan - White - Purple       */
/*                        21 = Program 1: Change 2: 6 Seconds hold, 0 Seconds change    */
/*                                                  Red - Green - Blue - Yellow -       */
/*                                                  Purple - Yellow - Cyan - Purple     */
/*                        22 = Program 2: Change 3: 3 Seconds hold, 3 Seconds change    */
/*                                                  Red - Green - Blue - Yellow         */
/*                                                  Puple - Yellow - Cyan - Purple      */
/*                        23 = Program 3: Change 4: 0 Seconds hold, 6 Second change     */
/*                                                  Red - Orange - Yellow - Light green */
/*                                                  Green - Cyan - Blue - Light Purple  */
/*                                                  Purple - Dark purple                */
/*                        24 = Program 4: Change 5: Spektrum 0 Sec. hold, 6 Sec. change */
/*                                                  Yellow - Red -  Purple - Blue       */
/*                                                  Cyan - Green                        */
/*                        25 = Program 5: Steps: Red Up/down, Green Up/Down,            */
/*                                               Blue Up/Down                           */
/*                        26 = Program 6: Color Flash 1: Green - Red - Blue             */
/*                                                     Yellow - Cyan - Purple           */
/*                                                     Green - Blue - Purple            */
/*                                                     Green - Orange - White           */
/*                                                     Cyan - Yellow - White            */
/*                        27 = Program 7: Color Flash 2: Yellow - Red - Cyan - Yellow   */
/*                                                       Cyan - Purple - Green - Blue   */
/*                                                       Purple - Green - Orange - White*/
/*                                                       Green - Yellow - Purple        */
/*                        28 = Program 8: Fireplace: Yellow Orange Up/Down              */
/*                        29 = Program 9: Thunderstorm: Blue flashing                   */
/*                                                                                      */
/*                        32-40: Dimmer values 0-8 for single colors                    */
/*                        48-51: Dimmer values 0-3 for programs                         */
/*                        64-73: Speed values 0-9 for programs                          */
/*                                                                                      */
/*                        128 = Power off                                               */
/*                        129 = Power on                                                */
/*                        130 = Pause                                                   */
/*                        131 = Play                                                   `*/
/*                        136 = Store value in storage position 1                       */
/*                        137 = Store value in storage position 2                       */
/*                        138 = Store value in storage position 3                       */
/*                        139 = Store value in storage position 4                       */
/*                        140 = Store value in storage position 5                       */
/*                        143 = Restore default values in storage 1-5                   */
/*                                                                                      */  
/* FLASH structure                                                                      */
/*                        0 = Actual program (1-13, 15-19, 20-29)                       */
/*                        1 = Color dimming factor (0-8) 0=no dimming, 8=max dimming    */
/*                        2 = Program dimming factor (0-3)                              */
/*                            0 = no dimming --> Stored value 0                         */
/*                            1 = dim step 1 --> Stored value 63                        */
/*                            2 = dim step 2 --> Stored value 126                       */
/*                            3 = dim step 3 --> Stored value 189                       */
/*                        3 = Program speed (0-9), 0=highest, 9=lowest                  */
/*                        10-13 Storage position 1                                      */
/*                              10 = Red color value                                    */
/*                              11 = Green color value                                  */
/*                              12 = Blue color value                                   */
/*                              13 = Color dimming (0-8) 0=no dimming, 8=max dimming    */
/*                        20-23 Storage position 2                                      */
/*                              20 = Red color value                                    */
/*                              21 = Green color value                                  */
/*                              22 = Blue color value                                   */
/*                              23 = Color dimming (0-8) 0=no dimming, 8=max dimming    */
/*                        30-33 Storage position 3                                      */
/*                              30 = Red color value                                    */
/*                              31 = Green color value                                  */
/*                              32 = Blue color value                                   */
/*                              33 = Color dimming (0-8) 0=no dimming, 8=max dimming    */
/*                        40-43 Storage position 4                                      */
/*                              40 = Red color value                                    */
/*                              41 = Green color value                                  */
/*                              42 = Blue color value                                   */
/*                              43 = Color dimming (0-8) 0=no dimming, 8=max dimming    */
/*                        50-53 Storage position 5                                      */
/*                              50 = Red color value                                    */
/*                              51 = Green color value                                  */
/*                              52 = Blue color value                                   */
/*                              53 = Color dimming (0-8) 0=no dimming, 8=max dimming    */
/*                                                                                      */
/* Display duing boot up:                                                               */
/*         Initialize SPIFFS:                                                           */
/*                    Failed: Fast blinking red 0,5 sec                                 */
/*                    Done: None                                                        */
/*         Read init file:                                                              */
/*                    Failed: Slower blinking red 1 sec                                 */
/*                    Done: None                                                        */
/*                                                                                      */
/*                                                                                      */
/****************************************************************************************/


/****************************************************************************************/
/* Libraries                                                                            */
/****************************************************************************************/
#include <EEPROM.h>                               /* Include EEPROM read/write */
#include <WiFi.h>                                 /* Include ESP32 WIFI library */
#include <ESPAsyncWebServer.h>                    /* Include ESP32 Async Web Server librarby */ 
#include "SPIFFS.h"                               /* Include ESP32 library to access files */
#include "FastLED.h"                              /* Include Fast LED NEO Pixel library */


/****************************************************************************************/
/* General FastLed settings                                                             */
/****************************************************************************************/
FASTLED_USING_NAMESPACE                           /* Definitions for ESP32 for Fastled */                       
#if defined(FASTLED_VERSION) && (FASTLED_VERSION < 3001000)
  #warning "Requires FastLED 3.1 or later; check github for latest code."
#endif


/****************************************************************************************/
/* Specific settings for FastLED library                                                */
/****************************************************************************************/
/* Definition for a non ESP32 Pico kit */
/* #define LED_DATA_PIN    12                        Led data pin definitions */
/* Definition for a ESP32 Pico kit */
#define LED_DATA_PIN    2                         /* IO2 = GPIO12 = Led data pin definitions */ 
#define LED_TYPE        WS2812B                   /* Led type definition */
#define LED_COLOR_ORDER RGB                       /* Led color order definition */
#define LED_NUM_LEDS    56                        /* Number of Leds */
#define LED_FRAMES_PER_SECOND  120                /* Constant to define frames per seconds */
CRGB leds[LED_NUM_LEDS];                          /* Reserve buffer for Led data */
#define FASTLED_SHOW_CORE 0                       /* Define the core to run FastLED.show() */
/* Task handles for use in the notifications */
static TaskHandle_t FAST_LED_SHOW_TASK_HANDLE = 0;    
static TaskHandle_t USER_TASK_HANDLE = 0;


/****************************************************************************************/
/* Definitions of hardware pins                                                         */
/****************************************************************************************/
/* Definition for a non ESP32 Pico kit */
/* const byte PUSH_BUTTON = 13;                      Push button reading pin */
/* Definition for a ESP32 Pico kit */
const byte PUSH_BUTTON = 15;                      /* IO15 = GPIO13 = Push button reading pin */


/****************************************************************************************/
/* Definitions for general settings                                                     */
/****************************************************************************************/
/* If this is uncommented here, status messages will be output via the serial interface, if commented out then disabled ! */
// #define _DEBUG_                                   
#define EEPROM_SIZE 60                            /* Define the number of bytes of FLASH to be used */


/****************************************************************************************/
/* Definitions for WIFI settings and handling                                           */
/****************************************************************************************/
char WLAN_SSID[21] = {0};                         /* String for WLAN_SSID */
char WLAN_PWD [21] = {0};                         /* String for password of WLAN */
IPAddress FIXED_IP;                               /* IP address for fixed IP */
IPAddress SUBNET;                                 /* Subnet mask */


/****************************************************************************************/
/* Definitions for Webserver                                                            */
/****************************************************************************************/
AsyncWebServer server(80);                        /* Create AsyncWebServer object on port 80 */


/****************************************************************************************/
/* Constants                                                                            */
/****************************************************************************************/
const byte NO_DIM_STEPS=200;                      /* Number of dim steps per sequence */
const char* SW_VERSION = "2.1";                   /* Actual SW version */
const byte RECEIVE_BUFFER_SIZE = 160;             /* 160 byte receiver buffer */
const byte TEMP_BUFFER_SIZE = 160;                /* 160 byte temporarly buffer */

/* Array for color values */
const char *NR_COLOR[] = { "White", "Warm white", "Yellow", "Orange", "Red", "Dark purple", "Purple", "Bright purple", "Dark blue", "Blue", "Cyan", "Bright green", "Green"};
/* Array for program values */
const char *NR_PROGRAM[] = { "Color change 1", "Color change 2", "Color change 3", "Color change 4", "Color change 5", "Steps", "Color flash 1", "Color flash 2", "Fireplace", "Thunderstorm"};
/* Array for storage positions */
const char *NR_STORAGE[] = { "Position 1", "Position 2", "Position 3", "Position 4", "Position 5"};       
/* Array for color brightness values */
const char *DIM_COLOR[] = { "100", "85", "70", "60", "50", "40", "30", "20", "10"};
/* Array for program brightness values */
const char *DIM_PROGRAM[] = { "100", "75", "50","25"};
/* Array for speed of program */
const char *SPEED_PROGRAM[] = { "100", "90", "80", "70", "60", "50", "40", "30", "20", "10"};

/* Sequenz array for program pointer */
/* This table is the pointer to the different programs */
/* for the table PROG_DATA, e.g. program 1 points to   */
/* 0, program 2 points to 61, ...                      */
const int PROG_POINTER[] = {0,61,102,143,194,225,346,497,648,749};

/* Sequenz array for dimming steps in percent */
/* e.g. 100 = 100%, 60 = 60%                  */
const int DIM_CONST[] = {100,60,30,15};

/* Sequenz array for dimming steps in pervcent */
/*e.g. 100 = 3 sec, 200 = 6 sec, 10 = 0,3 sec  */
const int SPEED_CONST[] = {6, 10, 25, 50, 75, 100, 200, 400, 800, 1600};

/* Sequenz array for program */
/* The dataset contains the following information */
/* The first value defines the number of steps    */
/* Then each data set consists of 5 values        */
/*     First value: Red value                     */
/*     Second value: Green value                  */
/*     Third value: Blue value                    */
/*     Forth value: Seconds for change            */
/*     Fifth value: Seconds for hold              */
/**************************************************/
const unsigned char PROG_DATA[] PROGMEM = {
12,
255,255,126,3,3,
255,0,0,3,3,
255,255,126,3,3,
0,255,0,3,3,
255,255,126,3,3,
0,0,255,3,3,
255,255,126,3,3,
255,255,0,3,3,
255,255,126,3,3,
0,255,255,3,3,
255,255,126,3,3,
255,0,255,3,3,
8,            
255,0,0,6,0,
0,255,0,6,0,
0,0,255,6,0,
255,255,0,6,0,
255,0,255,6,0,
255,255,0,6,0,
0,255,255,6,0,
255,0,255,6,0,
8,          
255,0,0,3,3,
0,255,0,3,3,
0,0,255,3,3,
252,252,0,3,3,
255,0,255,3,3,
255,255,0,3,3,
0,255,255,3,3,
255,0,255,3,3,
10,
255,0,0,6,0,
255,72,0,6,0,
255,255,0,6,0,
92,255,0,6,0,
0,255,0,6,0,
0,54,255,6,0,
0,0,255,6,0,
90,0,255,6,0,
255,0,255,6,0,
255,0,90,6,0,
6,      
255,255,0,6,0,
255,0,0,6,0,
255,0,255,6,0,
0,0,255,6,0,
0,255,255,6,0,
0,255,0,6,0,
24, 
0,0,0,4,0,   
18,0,0,4,2,
9,0,0,4,2,
36,0,0,4,2,
9,0,0,4,2,
92,0,0,4,2,
9,0,0,4,2,
255,0,0,4,2,
0,0,0,4,0,
0,18,0,4,2,
0,9,0,4,2,
0,36,0,4,2,
0,9,0,4,2,
0,92,0,4,2,
0,9,0,4,2,
0,255,0,4,2,
0,0,0,4,0,
0,0,18,4,2,
0,0,9,4,2,
0,0,36,4,2,
0,0,9,4,2,
0,0,92,4,2,
0,0,9,4,2,
0,0,255,4,3,
30,
0,252,0,0,3,
0,0,0,0,3,
252,0,0,0,3,
0,0,0,0,3,
0,0,252,0,3,
0,0,0,0,3,
252,252,0,0,3,
0,0,0,0,3,
0,252,252,0,3,
0,0,0,0,3,
252,0,252,0,3,
0,0,0,0,3,
0,252,0,0,3,
0,0,0,0,3,
0,0,252,0,3,
0,0,0,0,3,
252,0,252,0,3,
0,0,0,0,3,
0,252,0,0,3,
0,0,0,0,3,
252,35,0,0,3,
0,0,0,0,2,
252,252,252,0,3,
0,0,0,0,3,
0,252,252,0,3,
0,0,0,0,3,
252,252,0,0,3,
0,0,0,0,3,
252,252,252,0,3,
0,0,0,0,3,
30,
144,252,0,0,4,
0,0,0,0,2,
252,0,0,0,4,
0,0,0,0,2,
0,117,252,0,4,
0,0,0,0,2,
252,99,0,0,4,
0,0,0,0,2,
0,252,117,0,4,
0,0,0,0,2,
126,0,252,0,4,
0,0,0,0,2,
0,252,0,0,4,
0,0,0,0,2,
0,0,252,0,4,
0,0,0,0,2,
252,0,63,0,4,
0,0,0,0,2,
0,252,108,0,4,
0,0,0,0,2,
252,18,0,0,4,
0,0,0,0,2,
252,252,252,0,4,
0,0,0,0,2,
0,252,144,0,4,
0,0,0,0,2,
252,144,16,0,4,
0,0,0,0,2,
36,0,252,0,4,
0,0,0,0,2,
20,
252,36,0,2,3,
252,18,0,0,2,
252,36,0,0,2,
252,45,0,3,3,
252,36,0,0,3,
252,18,0,2,0,
252,45,0,4,2,
252,18,0,0,2,
252,45,0,6,4,
252,54,0,0,2,
252,45,0,2,3,
252,36,0,0,3,
252,18,0,0,2,
252,36,0,3,2,
252,18,0,0,2,
252,45,0,0,3,
252,18,0,3,0,
252,36,0,0,2,
252,45,0,3,0,
252,18,0,0,2,
18,
0,0,0,0,3,
0,27,171,0,1,
0,0,0,0,2,
0,117,252,0,2,
0,252,252,4,1,
0,252,117,2,1,
0,54,252,0,3,
0,0,0,0,3,
0,0,63,0,1,
0,0,126,0,1,
0,0,189,0,1,
0,54,252,0,1,
0,0,0,0,2,
0,63,126,2,1,
0,117,252,2,1,
0,252,252,2,1,
0,252,117,2,1,
0,126,63,2,1
};


/****************************************************************************************/
/* Variables                                                                            */
/****************************************************************************************/
boolean POWER_OFF;                                /* True when standby activated */
boolean POWER_WEB_OFF;                            /* True when standby activated to show correct information on web page */
boolean TEST_COLOR;                               /* False = No test color, True = Test color */
boolean SUB_MENU_FLAG;                            /* Flag for sub menu; False = Display storage settings with 0, true = Load actual storage settings after recall */
boolean KEY_PRESSED;                              /* Flag for key pressed for debouncing delay: True = Key pressed, False = No key pressed */
boolean DISPLAY_LOCK;                             /* Flag to lock display of Neopixel e.g. during update of website, False = No lock, True = Lock */
byte SELECT;                                      /* Select value */
byte ACT_PROGRAM;                                 /* Actual program */
byte ACT_COLOR;                                   /* Actual color */
byte COLOR_DIM_STEP;                              /* Actual color dimming factor */
byte PROG_DIM_STEP;                               /* Actual program dimming factor */
byte PROG_SPEED;                                  /* Program speed */
byte MAX_RED_COLOR;                               /* Maximum dimmer value red color */
byte MAX_GREEN_COLOR;                             /* Maximum dimmer value green color */
byte MAX_BLUE_COLOR;                              /* Maximum dimmer value blue color */
byte RED_DIM;                                     /* Red color dimmer value */
byte GREEN_DIM;                                   /* Green color dimmer value */
byte BLUE_DIM;                                    /* Blue color dimmer value */
byte RED_COLOR;                                   /* Red color value */
byte GREEN_COLOR;                                 /* Green color value */
byte BLUE_COLOR;                                  /* Blue color value */
byte STORE_RED_COLOR;                             /* Red color value of selected storage position */
byte STORE_GREEN_COLOR;                           /* Green color value of selected storage position */
byte STORE_BLUE_COLOR;                            /* Blue color value of selected storage position */
byte STORE_DIM_VALUE;                             /* Dimming value of selected storage position */
uint16_t LOOP_COUNTER;                            /* Universal loop counter */
String RECEIVE_BUFFER = "";                       /* String to receive data */
String TEMP_BUFFER = "";                          /* String for string manipulation */


/****************************************************************************************/
/* Setup procedure                                                                      */
/****************************************************************************************/
void setup() 
{ 
  TEMP_BUFFER.reserve(TEMP_BUFFER_SIZE);          /* Reserve byte buffer for packet buffer string */ 
  RECEIVE_BUFFER.reserve (RECEIVE_BUFFER_SIZE);   /* Reserve byte buffer for receive buffer */
  
  /* Initialize FLASH size */
  EEPROM.begin(EEPROM_SIZE);                      /* Initialize FLASH with predefined size */

  /* Initialise pins */
  pinMode(PUSH_BUTTON,INPUT_PULLUP);              /* Define push button as intput with pullup */

  /* Init serial interface for debug information */
  Serial.begin(115200);                           /* Start serial interface for terminal communication at 115000 Baud */
  /* Display welcome message with initialisation information */
  Serial.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  Serial.println ("*********************************************************");
  Serial.println ("* Neopixel bulb controller for ESP32                    *");
  Serial.println ("*                                                       *");
  Serial.print   ("* V");
  Serial.print   (SW_VERSION);
  Serial.println ("                                      @PT 2020    *");   
  Serial.println ("*********************************************************\n"); 

  #ifdef _DEBUG_
    Serial.println("\n\n\n"); 
    Serial.println ("Debug is enabled...");
    Serial.println(); 
  #else
    Serial.println("\n\n\n"); 
    Serial.println ("Debug is disabled...");
    Serial.println ("Remove commented out line #define _DEBUG_ to enable!");
    Serial.println();     
  # endif  

  DISPLAY_LOCK = false;                           /* Set display lock false */
  /* Read init parameters from FLASH */
  if (EEPROM.read(0)==255)                        /* FLASH never written before? --> First intialisation */
  {
    SUB_INIT_FLASH ();                            /* Call subroutine to initialize FLASH with default values */
  }

  /* Read out FLASH values */
  #ifdef _DEBUG_
    SUB_DISPLAY_FLASH ();                         /* Dump all read FLASH values in debug mode */
  # endif 

  /* Step 1: Initialize Neopixel library and switch off LEDs */
  #ifdef _DEBUG_
    Serial.println();
    Serial.print ("Step 1: Init Neopixel interface and switch channels off: ");
  # endif
  /* Initialize FastLED with the right configuration for the strip */
  FastLED.addLeds<LED_TYPE,LED_DATA_PIN,LED_COLOR_ORDER>(leds, LED_NUM_LEDS).setCorrection(TypicalLEDStrip);
  /* Create the FastLED show task */
  xTaskCreatePinnedToCore(SUB_FAST_LED_SHOW_TASK, "SUB_FAST_LED_SHOW_TASK", 2048, NULL, 2, &FAST_LED_SHOW_TASK_HANDLE, FASTLED_SHOW_CORE);
  delay (500);                                    /* Wait 500ms */
  SUB_NEOPIXEL_SEND (0,0,0);                      /* Switch all LEDs off */ 
  delay (500);                                    /* Wait 5000ms */
  FastLED.delay(1000/LED_FRAMES_PER_SECOND);      /* This sets the frame update delay */   
  #ifdef _DEBUG_
    Serial.println("Done");                       /* Display success message */
    Serial.println();
  # endif

   /* Step 2: Initialize SPIFFS */
  #ifdef _DEBUG_
    Serial.println ("Step 2: Initialise file access and read Init.txt...");
    Serial.print("  Initialize file access: ");
  # endif 
  if(!SPIFFS.begin(true))                         /* Initialisation error ? */
  {  
    #ifdef _DEBUG_                                          
      Serial.println("Failed!");
    # endif
    do                                            /* Endless loop until manual reset */
    {
      SUB_NEOPIXEL_SEND (255, 0, 0);              /* Switch all pixels to red color */
      delay (500);                                /* 500ms delay */
      SUB_NEOPIXEL_SEND (0, 0, 0);                /* Switch off all pixels */
      delay (500);                                /* 500ms delay */      
    }
    while (1);  
  }
  else                                            /* No error ? */
  { 
    #ifdef _DEBUG_                                            
      Serial.println("Done");                     /* Display success message */
    # endif  
    SUB_READ_INIT_FILE();                         /* Call subroutine to read init file */
  }  

  /* Step 3: Create Wifi Access point */
  #ifdef _DEBUG_
    Serial.println();
    Serial.print ("Step 3: Create Wifi Access point with SSID '");
    Serial.print (WLAN_SSID);
    Serial.print ("' and password '");
    Serial.print (WLAN_PWD);
    Serial.print ("': ");
  # endif 
  WiFi.mode(WIFI_AP);                             /* Set to Access point mode */
  WiFi.softAP(WLAN_SSID, WLAN_PWD);               /* Initialize the Wifi Soft Access Point */
  delay(100);                                     /* Wait 100ms for AP start */
  WiFi.softAPConfig(FIXED_IP, FIXED_IP, SUBNET);  /* Configure IP address for access point */
  #ifdef _DEBUG_
    Serial.println("Done");                       /* Display success message */
    Serial.print ("  INFO: Connect to IP-Address '");
    Serial.print (WiFi.softAPIP());
    Serial.println ("' to display webpage!");
  # endif 
  
  /* Step 4: Define routes for answers from web page */
  #ifdef _DEBUG_
    Serial.println();
    Serial.print ("Step 4: Define routes for answers from web page: ");
  # endif 
  SUB_DEFINE_ROUTES ();                           /* Call subroutine to define routes for HTML equests */
  server.begin();                                 /* Start server */
  #ifdef _DEBUG_
    Serial.println("Done");                       /* Display success message */
  # endif 

  /* Initialize variables */
  POWER_OFF = true;                               /* Keep device switched off */
  POWER_WEB_OFF = true;                           /* Web display switched off */
  TEST_COLOR    = false;                          /* Set test color flag false */
  KEY_PRESSED   = false;                          /* Set key pressed flag false */
  SELECT = 255;                                   /* Delete key buffer */
  ACT_PROGRAM    = EEPROM.read (0);               /* Read actual program out of FLASH */
  COLOR_DIM_STEP = EEPROM.read (1);               /* Read actual color dimming value out of FLASH */
  PROG_DIM_STEP  = EEPROM.read (2);               /* Read actual program dimming value out of FLASH */
  PROG_SPEED     = EEPROM.read (3);               /* Read program speed out of FLASH */

  Serial.println();
  Serial.println ("Initialisation finished! Ready to receive requests ...");
  Serial.println ();
}

  
/****************************************************************************************/
/* Main Loop                                                                            */
/****************************************************************************************/
void loop() 
{
  /****************************************************************/
  /* Check to test color                                          */
  /****************************************************************/ 
  if (TEST_COLOR == true)                         /* Test color flag true ? */
  {  
    #ifdef _DEBUG_
      Serial.println ();
      Serial.println ("Set test color ...");
    #endif
    SUB_TEST_VALUE();                             /* Call subroutine to set test value */
    TEST_COLOR = false;                           /* Set test color false */
  }
  
  /****************************************************************/
  /* Check for push button pressed                                */
  /****************************************************************/
  if (digitalRead (PUSH_BUTTON) == 0)             /* Key pressed ? */
  {
    #ifdef _DEBUG_
      Serial.println ();
      Serial.println ("Push button detected!");
    #endif
    KEY_PRESSED = true;                           /* Set key pressed flag true */
    if (POWER_OFF == true)                        /* Device is not powered on */
    {
      SELECT=129;                                 /* Set select to 129 = Power On */                               
    }
    else                                          /* Device is powered on */
    { 
      SELECT=128;                                 /* Set select to 128 = Power Off */                               
    }
  }

  /****************************************************************/
  /* Power On Selected                                            */
  /****************************************************************/
  if ((POWER_OFF == true) && (SELECT==129))
  {
    do                                            /* Wait until webpage has loaded */
    {
      delay (10);                                 /* Delaytime to enable other tasks to be performed */
    }
    while (DISPLAY_LOCK == true);
    delay (300);                                  /* 300 ms debouncing delay */
    SUB_POWER_ON();                               /* Call subroutine for power on */
  }

  /****************************************************************/
  /* Power Off Selected                                           */
  /****************************************************************/
  if ((POWER_OFF == false) && (SELECT==128))
  {
    do                                            /* Wait until webpage has loaded */
    {
      delay (10);                                 /* Delaytime to enable other tasks to be performed */
    }
    while (DISPLAY_LOCK == true);
    delay (300);                                  /* 300 ms debouncing delay */
    SUB_POWER_OFF();                              /* Call subroutine for power off */
  }

  /****************************************************************/
  /* Select 1-13: Single color 1-13 selected                      */
  /****************************************************************/
  if ((SELECT >=1) && (SELECT <= 13))
  {
    if (POWER_OFF == true)                        /* Switched on from power off state */
    {
      delay (300);                                  /* 300 ms debouncing delay */
      POWER_OFF = false;                          /* Set power off false */
      POWER_WEB_OFF = false;                      /* Set power off false for web */
      COLOR_DIM_STEP = EEPROM.read (1);           /* Read actual color dimming value out of FLASH */
    }
    do                                            /* Wait until webpage has loaded */
    {
      delay (10);                                 /* Delaytime to enable other tasks to be performed */
    }
    while (DISPLAY_LOCK == true);
    SUB_MAIN_COLOR ();                            /* Call subroutine for single color with restore */
  } 

  /****************************************************************/
  /* Program 15-19 (Recall 1-5) selected                          */
  /****************************************************************/
  if ((SELECT >=15) && (SELECT <= 19))
  {
    if (POWER_OFF == true)                        /* Switched on from power off state */
    {
      delay (300);                                  /* 300 ms debouncing delay */
      POWER_OFF = false;                          /* Set power off false */
      POWER_WEB_OFF = false;                      /* Set power off false for web */
    }    
    do                                            /* Wait until webpage has loaded */
    {
      delay (10);                                 /* Delaytime to enable other tasks to be performed */
    }
    while (DISPLAY_LOCK == true);
    SUB_RCL_STORAGE();                            /* Call subroutine for single storage */  
  } 

  /****************************************************************/
  /* Program 20-29 (Program 1-5) selected                         */
  /****************************************************************/
  if ((SELECT >=20) && (SELECT <= 29))
  {
    if (POWER_OFF == true)                        /* Switched on from power off state */
    {
      POWER_OFF = false;                          /* Set power off false */
      POWER_WEB_OFF = false;                      /* Set power off false for web */
      PROG_DIM_STEP = EEPROM.read (2);            /* Read actual color dimming value out of FLASH */
      PROG_SPEED    = EEPROM.read (3);            /* Read program speed out of FLASH */
    }
    do                                            /* Wait until webpage has loaded */
    {
      delay (10);                                 /* Delaytime to enable other tasks to be performed */
    }
    while (DISPLAY_LOCK == true);
    SUB_PROG_SEL ();                              /* Call subroutine for animated programs */  
  }
  
  /****************************************************************/
  /* Program 136-140 (Store 1-5) selected                         */
  /****************************************************************/
  if ((SELECT >=136) && (SELECT <= 140) && (POWER_OFF == false))
  {
    do                                            /* Wait until webpage has loaded */
    {
      delay (10);                                 /* Delaytime to enable other tasks to be performed */
    }
    while (DISPLAY_LOCK == true);
    SUB_SET_STORAGE();                            /* Call subroutine for set storage */   
  }

  /****************************************************************/
  /* Program 143 restore memory default selected                  */
  /****************************************************************/
  if (SELECT == 143)                              /* Restore color storage positions 1-5 */
  { 
    #ifdef _DEBUG_
      Serial.println ("");
      Serial.println ("Restoring storage positions 1-5 ... ");
    # endif      
    EEPROM.write(10, 144);                        /* Storage position 1: Red color value */ 
    EEPROM.write(11, 72);                         /* Storage position 1: Green color value */ 
    EEPROM.write(12, 18);                         /* Storage position 1: Blue color value */ 
    EEPROM.write(13,0);                           /* Storage position 1: Dimming factor = 0 */  
    EEPROM.write(20, 252);                        /* Storage position 2: Red color value */
    EEPROM.write(21, 45);                         /* Storage position 2: Green color value */
    EEPROM.write(22, 0);                          /* Storage position 2: Blue color value */
    EEPROM.write(23,0);                           /* Storage position 2: Dimming factor = 0 */ 
    EEPROM.write(30, 72);                         /* Storage position 3: Red color value */
    EEPROM.write(31, 45);                         /* Storage position 3: Green color value */
    EEPROM.write(32, 0);                          /* Storage position 3: Blue color value */
    EEPROM.write(33,0);                           /* Storage position 3: Dimming factor = 0 */ 
    EEPROM.write(40, 112);                        /* Storage position 4: Red color value */
    EEPROM.write(41, 0);                          /* Storage position 4: Green color value */
    EEPROM.write(42, 56);                         /* Storage position 4: Blue color value */
    EEPROM.write(43,0);                           /* Storage position 4: Dimming factor = 0 */ 
    EEPROM.write(50, 27);                         /* Storage position 5: Red color value */
    EEPROM.write(51, 99);                         /* Storage position 5: Green color value */
    EEPROM.write(52, 252);                        /* Storage position 5: Blue color value */
    EEPROM.write(53,0);                           /* Storage position 5: Dimming factor = 0 */ 
    EEPROM.commit();                              /* Store values permanently in EEPROM */ 
    delay (100);                                  /* 100ms delay */
    SELECT = 255;                                 /* Set command to 255 */
  }
} 
